#include <iostream> 
#include <list> 
#include <cstring>
#include <iterator> 
#include <unordered_map>
#include "cache.h"
using namespace std;
/* 
 * The idea behind cache implementation :
 * To implement LRU, Write-back, fully associative cache
 * MAX Cache size - 32 key-value pairs
 * Each cache node will consist of 1. key 2. value 3. dirty bit
 * Algo :
 * 1. Search in cache :
 * Use unordered_map to see if a key is present in the cache and if it is present, this operation
 * should return 1
 * 2. Insert in cache :
 * If cache is full, call LRU
 * If cache has is not full, replace the frame with key-value pair to be inserted 
 * 3. LRU cache :
 * Keep a pointer on a frame after the frame last replaced 
 * Move the pointer circularly, if a node has referenceBit = '0', then select this key-value for 
 * replacement
 * Use bitmap to identify the file in which key is present
 * (Use locality to evict more than one key-value pair?)
 * And read the key-value pair in array element
 * 
 * Data Structures : 1. List of cacheNode - 32 keys
 *                   2. Unordered map - <key, array index>
 *  
 */

#define MAX_CACHE_SIZE 32

list<cacheNode> cache;
unordered_map<char *, list <cacheNode> :: iterator> hashTable;

int search(char* key) {
    unordered_map<char *, list <cacheNode> :: iterator> :: iterator it = hashTable.find(key);
    if(it == hashTable.end()) {
        return 0;
    }
    else {
        list <cacheNode> :: iterator cnode_it = it->second;
        cache.push_front(*cnode_it);
        cache.erase(cnode_it);
        return 1;
    }
}

void lru(char* key, char *value) {

}

/*
 * Assumption here is that before calling insert, key-value pair does not already exist in cache
 */
void insert(char* key, char* value) {
    cacheNode cnode;
    if(cache.size() == MAX_CACHE_SIZE) {
        lru(key, value);
    }
    else {
        cnode.key = (char *)malloc(strlen(key) + 1);
        strcpy(cnode.key, key);
        cnode.value = (char *)malloc(strlen(value) + 1);
        strcpy(cnode.value, value);
        cnode.dirtyBit = '0';
        cache.push_front(cnode);
        hashTable[key] = cache.begin();

    }
}

/* 
 * Testing code
 */
int main() {
    char key1[] = "A";
    char key2[] = "B";
    char key3[] = "C";
    char key4[] = "D";
    char value[] = "1";
    insert(key1, value);
    insert(key2, value);
    insert(key3, value);
    printf("%d\n", search(key1));
    printf("%d\n", search(key2));
    printf("%d\n", search(key3));
    printf("%d\n", search(key4));
}
